-----------------------------------------------------------------------------------------------
-- Client Lua Script for CharacterInfoPlus
-- Copyright (c) NCsoft. All rights reserved
-----------------------------------------------------------------------------------------------
 
require "Window"
 
-----------------------------------------------------------------------------------------------
-- CharacterInfoPlus Module Definition
-----------------------------------------------------------------------------------------------
local CharacterInfoPlus = {} 
 
-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
CharacterInfoPlus.CharacterInfoLines = {}
CharacterInfoPlus.LineQueue = {}
 
-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function CharacterInfoPlus:new(o)
    o = o or {}
    setmetatable(o, self)
    self.__index = self 
	
	self.CharacterInfoLines = {}
	self.LineQueue = {}
	
    return o
end

function CharacterInfoPlus:Init()
	Apollo.RegisterAddon(self, true)
end
 

-----------------------------------------------------------------------------------------------
-- CharacterInfoPlus OnLoad
-----------------------------------------------------------------------------------------------
function CharacterInfoPlus:OnLoad()
    -- load our form file
	self.xmlDoc = XmlDoc.CreateFromFile("CharacterInfoPlus.xml")
	
	Apollo.GetPackage("Gemini:Hook-1.0").tPackage:Embed(self)
	self.config = Apollo.GetPackage("Gemini:DB-1.0").tPackage:New(self)
	
	local charAddons = {
		"RealAP",
		"Character"
	}
	for k,v in pairs(charAddons) do
		self.CharacterAddon = Apollo.GetAddon(v)
		if self.CharacterAddon then break end
	end
	
	self:PostHook(self.CharacterAddon, "OnToggleCharacterWindow")
	self:PostHook(self.CharacterAddon, "OnPersonaUpdateCharacterStats")
	
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
end

function CharacterInfoPlus:OnDocLoaded()
	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
		self.wndSettings = Apollo.LoadForm(self.xmlDoc, "SettingsWindow", nil, self)
		self:initLocation()
		self.wndSettings:Show(false, true)
		
		Apollo.RegisterSlashCommand("cip", "OnConfigure", self)
	end
end

function CharacterInfoPlus:initLocation()
	if self.config.char.Position ~= nil then
		self.wndSettings:SetAnchorOffsets(self.config.char.Position.left, self.config.char.Position.top, self.config.char.Position.right, self.config.char.Position.bottom)
	else 
		self.config.char.Position = {}
		self.config.char.Position.left, self.config.char.Position.top, self.config.char.Position.right, self.config.char.Position.bottom = self.wndSettings:GetAnchorOffsets()
	end
end


-----------------------------------------------------------------------------------------------
-- CharacterInfoPlus Functions
-----------------------------------------------------------------------------------------------
function CharacterInfoPlus:OnToggleCharacterWindow()
	self:InitCharacterWindow()
end

function CharacterInfoPlus:InitCharacterWindow()
	if self.wndData == nil then
	
		local charAddons = {
			"RealAP",
			"Character"
		}
		for k,v in pairs(charAddons) do
			self.CharacterAddon = Apollo.GetAddon(v)
			if self.CharacterAddon then break end
		end
	
		self.wndData = Apollo.LoadForm(self.xmlDoc, "CharacterMoreInfosForm", self.CharacterAddon.wndCharacter, self)
	end
	
	self:ConsumeQueue()
end

function CharacterInfoPlus:ConsumeQueue()
	for addonName, addonLines in pairs(self.LineQueue) do
		for k,v in self:spairs(addonLines) do
			self:ExecuteQueue(v)
		end
	end

	--SendVarToRover("Char", self.CharacterInfoLines)
end

function CharacterInfoPlus:RegiterLine(addonName, title)
	if(self.LineQueue[addonName] == nil) then
		self.LineQueue[addonName] = {}
	end
	if(self.LineQueue[addonName][title] == nil) then
		self.LineQueue[addonName][title] = {}
	end
	self.LineQueue[addonName][title].title = title
	self.LineQueue[addonName][title].addonName = addonName
	--SendVarToRover("queue - " .. addonName .. " - " .. title, self.LineQueue[addonName][title])
end

function CharacterInfoPlus:AddCharacterInformation(addonName, title, functionForValue, functionForColor, data, tooltipContent)
	if self.wndData == nil then
		if self:PreInitLine(addonName, title, functionForValue, functionForColor, data, tooltipContent) == nil then 
			Print("<CharacterInfoPlus.AddLine> DEBUG HELP: failed to preInit the line.")
			return nil 
		end
		self:RegiterLine(addonName, title)
		return self.CharacterInfoLines[addonName][title]
	else
		return self:ExecuteLine(addonName, title, functionForValue, functionForColor, data, tooltipContent)
	end
end

function CharacterInfoPlus:ExecuteQueue(queueElement)
	local t = self:ExecuteLine(queueElement.addonName, queueElement.title, nil, nil, nil, nil)
	queueElement = nil
	return t
end

function CharacterInfoPlus:ExecuteLine(addonName, title, functionForValue, functionForColor, data, tooltipContent)
	if self.CharacterInfoLines[addonName] ~= nil and self.CharacterInfoLines[addonName][title] ~= nil and self.CharacterInfoLines[addonName][title].complete then 
		return self.CharacterInfoLines[addonName][title]
	end
	if self.CharacterInfoLines[addonName] == nil then
		self.CharacterInfoLines[addonName] = {}
	end
	
	if self.CharacterInfoLines[addonName][title] == nil then
		if self:PreInitLine(addonName, title, functionForValue, functionForColor, data, tooltipContent) == nil then 
			Print("<CharacterInfoPlus.Execute> DEBUG HELP: failed to preInit the line.")
			return nil 
		end
	end
	
	local container = self.wndData:FindChild("InfosContainer")
	self.CharacterInfoLines[addonName][title].wnd = Apollo.LoadForm(self.xmlDoc, "GenericScoreCharacterLine", container , self)
	local wnddata = self.CharacterInfoLines[addonName][title].wnd:GetData()
	if wnddata == nil then wnddata = {} end
	wnddata.TITLE = title
	wnddata.ADDONNAME = addonName
	self.CharacterInfoLines[addonName][title].wnd:SetData(wnddata)
	self.CharacterInfoLines[addonName][title].wnd:FindChild("Title"):SetText(title .. ": ")
	self.CharacterInfoLines[addonName][title].wnd:Show(self.config.char.CharacterInfoLines[addonName][title].shown, true)
	self:UpdateCharacterInformation(self.CharacterInfoLines[addonName][title])
	self.CharacterInfoLines[addonName][title].complete = true
	container:ArrangeChildrenVert()
	
	return self.CharacterInfoLines[addonName][title]
end

function CharacterInfoPlus:PreInitLine(addonName, title, functionForValue, functionForColor, data, tooltipContent)
	
	if functionForValue == nil or type(functionForValue) ~= 'function' then 
		Print("<CharacterInfoPlus> DEBUG HELP: the update function is not a function type.")
		return nil 
	end
	if functionForColor ~= nil and type(functionForValue) ~= 'function' then 
		Print("<CharacterInfoPlus> DEBUG HELP: the function to update color is not nil and not a function")
		return nil 
	end
	if title == nil or type(title) ~= 'string' then 
		Print("<CharacterInfoPlus> DEBUG HELP: The title is not a string")
		return nil 
	end
	if addonName == nil or type(addonName) ~= 'string' then 
		Print("<CharacterInfoPlus> DEBUG HELP: the addon name is not a string")
		return nil 
	end
	if tooltipContent ~= nil and type(tooltipContent) ~= 'string' then 
		Print("<CharacterInfoPlus> DEBUG HELP: The tooltip is not nil but not a string")
		return nil 
	end
	
	if self.CharacterInfoLines[addonName] == nil then
		self.CharacterInfoLines[addonName] = {}
	end
	
	if self.config.char.CharacterInfoLines == nil then
		self.config.char.CharacterInfoLines = {}
	end
	
	if self.config.char.CharacterInfoLines[addonName] == nil then
		self.config.char.CharacterInfoLines[addonName] = {}
	end
	
	if self.config.char.CharacterInfoLines[addonName][title] == nil then
		self.config.char.CharacterInfoLines[addonName][title] = {}
		self.config.char.CharacterInfoLines[addonName][title].shown = true
	end
	
	if self.CharacterInfoLines[addonName][title] == nil then
		self.CharacterInfoLines[addonName][title] = {}
		self.CharacterInfoLines[addonName][title].title = title
		self.CharacterInfoLines[addonName][title].addonName = addonName
		self.CharacterInfoLines[addonName][title].update = functionForValue
		self.CharacterInfoLines[addonName][title].updateColor = functionForColor
		self.CharacterInfoLines[addonName][title].data = data
		self.CharacterInfoLines[addonName][title].forceUpdate = function() self:UpdateCharacterInformation(self.CharacterInfoLines[addonName][title]) end
		self.CharacterInfoLines[addonName][title].tooltipContent = tooltipContent
		self.CharacterInfoLines[addonName][title].complete = false
	end
	--SendVarToRover(addonName .. " - " .. title, self.CharacterInfoLines[addonName][title])
	return self.CharacterInfoLines[addonName][title]
end

function CharacterInfoPlus:CustomGenerateTooltip( wndHandler, wndControl, eToolTipType, x, y )

	wndControl:SetTooltipDoc(nil)
	local title = wndHandler:GetData().TITLE
	local addonName = wndHandler:GetData().ADDONNAME
	local xml = XmlDoc.new()
	xml:StartTooltip(Tooltip.TooltipWidth)
	local content = self.CharacterInfoLines[addonName][title].tooltipContent
	if content ~= nil then 
		xml:AddLine("[" .. addonName .. "] " .. content)
		wndControl:SetTooltipDoc(xml)
	end
end

function CharacterInfoPlus:RemoveCharacterInformation(addonName, title)
	if self.CharacterInfoLines[addonName] == nil or self.CharacterInfoLines[addonName][title] == nil then return end
	
	if self.wndData == nil then
		local charAddons = {
			"RealAP",
			"Character"
		}
		for k,v in pairs(charAddons) do
			self.CharacterAddon = Apollo.GetAddon(v)
			if self.CharacterAddon then break end
		end
	
		self.wndData = Apollo.LoadForm(self.xmlDoc, "CharacterMoreInfosForm", self.CharacterAddon.wndCharacter, self)
	end
	
	local container = self.wndData:FindChild("InfosContainer")
	self.CharacterInfoLines[addonName][title].wnd:Destroy()
	self.CharacterInfoLines[addonName][title] = nil
	
	container:ArrangeChildrenVert()
end

function CharacterInfoPlus:UpdateCharacterInformation(charInfo)
	if charInfo.wnd == nil then return false end
	if not charInfo.wnd:IsShown() then return false end
	
	charInfo.wnd:FindChild("Value"):SetText(charInfo:update())
	if charInfo.updateColor ~= nil then
		charInfo.wnd:FindChild("Value"):SetTextColor(charInfo:updateColor())
	end
	return true
end

function CharacterInfoPlus:UpdateAll()
	for addonName, addonLines in pairs(self.CharacterInfoLines) do
		for k,v in self:spairs(addonLines) do
			self:UpdateCharacterInformation(v)
		end
	end
end

function CharacterInfoPlus:OnPersonaUpdateCharacterStats()
	self:UpdateAll()
end

function CharacterInfoPlus:onToggleMoreScoresWindow( wndHandler, wndControl, eMouseButton )
	wndControl:GetParent():FindChild("MoreInfoRegion"):Show(wndControl:IsChecked())
end

function CharacterInfoPlus:ClearContainer(container)
	local kids = container:GetChildren()
	container:SetData(nil)
	for k,v in pairs(kids) do
		v:Destroy()
	end
end

---------------------------------------------------------------------------------------------------
-- Settings Functions
---------------------------------------------------------------------------------------------------

function CharacterInfoPlus:OnSettingsWindowUpdateSize( wndHandler, wndControl )
	self.config.char.Position.left, self.config.char.Position.top, self.config.char.Position.right, self.config.char.Position.bottom =  self.wndSettings:GetAnchorOffsets()
end

function CharacterInfoPlus:OnWindowSettingsClose( wndHandler, wndControl, eMouseButton )
	self.wndSettings:Show(false, true)
end

function CharacterInfoPlus:OnConfigure()
	self:ShowSettingsWindow()
end


function CharacterInfoPlus:spairs(t, order)
    -- collect the keys
    local keys = {}
    for k in pairs(t) do keys[#keys+1] = k end

    -- if order function given, sort by it by passing the table and keys a, b,
    -- otherwise just sort the keys 
    if order then
        table.sort(keys, function(a,b) return order(t, a, b) end)
    else
        table.sort(keys)
    end

    -- return the iterator function
    local i = 0
    return function()
        i = i + 1
        if keys[i] then
            return keys[i], t[keys[i]]
        end
    end
end

function CharacterInfoPlus:ShowSettingsWindow()
	self:InitCharacterWindow()
	self:ClearContainer(self.wndSettings:FindChild("LinesContainer"))
	table.sort(self.CharacterInfoLines, sort_func )
	
	for addonName, addonLines in pairs(self.CharacterInfoLines) do
		for k,v in self:spairs(addonLines) do
			self:AddCharacterSettingsLine(addonName, k)
		end
		--for k,v in pairs(addonLines) do
			--self:AddCharacterSettingsLine(addonName, k)
		--end
	end
	
	self.wndSettings:FindChild("LinesContainer"):ArrangeChildrenVert()

	self.wndSettings:Show(true, true)
end

function CharacterInfoPlus:AddCharacterSettingsLine(addonName, LineTitle)
	local line = Apollo.LoadForm(self.xmlDoc, "CharacterLineSettings", self.wndSettings:FindChild("LinesContainer"), self)
	line:FindChild("AddonName"):SetText(addonName)
	line:FindChild("LineTitle"):SetText(LineTitle)
	local data = line:FindChild("HideBtn"):GetData()
	if data == nil then data = {} end
	data.title = LineTitle
	data.addonName = addonName
	line:FindChild("HideBtn"):SetData(data)
	line:FindChild("HideBtn"):SetCheck(not self.config.char.CharacterInfoLines[addonName][LineTitle].shown)
end

---------------------------------------------------------------------------------------------------
-- CharacterLineSettings Functions
---------------------------------------------------------------------------------------------------

function CharacterInfoPlus:OnCharacterSettingsLineShownChange( wndHandler, wndControl, eMouseButton )
	local data = wndControl:GetData()
	local title = data.title
	local addonName = data.addonName
	self.config.char.CharacterInfoLines[addonName][title].shown = not wndControl:IsChecked()
	self.CharacterInfoLines[addonName][title].wnd:Show(self.config.char.CharacterInfoLines[addonName][title].shown, true)
	self:UpdateCharacterInformation(self.CharacterInfoLines[addonName][title])
	self.wndData:FindChild("InfosContainer"):ArrangeChildrenVert()
end

-----------------------------------------------------------------------------------------------
-- CharacterInfoPlus Instance
-----------------------------------------------------------------------------------------------
local CharacterInfoPlusInst = CharacterInfoPlus:new()
CharacterInfoPlusInst:Init()

function AddCharacterInformationLine(addonName, title, functionToGetValue, funtionToGetColor, data, tooltipContent)
	return CharacterInfoPlusInst:AddCharacterInformation(addonName, title, functionToGetValue, funtionToGetColor, data, tooltipContent)
end

function RemoveCharacterInformationLine(addonName, title)
	CharacterInfoPlusInst:RemoveCharacterInformation(addonName, title)
end